# Copyright 2017-2018 Aaron C. Prunty
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#        
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

import os
import sys
import numpy as np
from vezda.Tikhonov import Tikhonov
from vezda.FundamentalSolutions import FundSol
from scipy.fftpack import fft, ifft, fftfreq
#from scipy.optimize import fsolve
from scipy.linalg import norm
from tqdm import trange
from time import sleep
from pathlib import Path

sys.path.append(os.getcwd())
import pulseFun

def timeShift(data, tau, dt):
    '''
    Apply the time-shift parameter 'tau' to the test
    functions in 'data' using FFT.
    
    'data' is a 3D array with time on axis=1
    'tau' is a constant representing the time shift
    'dt' is the length of the time step (used to generate the discretized frequency bins)
    '''
    
    # Get the shape of the data array
    # Nr : number of receivers
    # Nt : number of time samples
    # Ns : number of sources
    Nr, Nt, Ns = data.shape
    
    # FFT the 'data' array into the frequency domain along the time axis=1
    # Pad the time axis with zeros to length 2 * Nt - 1
    freqData = fft(data, n=2*Nt-1, axis=1)
    
    # Set up the phase vector e^(-i * omega * tau)
    iomega = 2j * np.pi * fftfreq(2*Nt-1, dt)
    phase = np.exp(-iomega * tau)
    
    # Apply time shift in the frequency domain (element-wise array multiplication)
    # and inverse FFT back into the time domain (keeping only the real part)
    shiftedData = ifft(freqData * phase[None, :, None], axis=1).real
    
    return shiftedData[:, :Nt, :]

def solver(medium, alpha, tau):    
    # Read in the computed singular-value decomposition of the near-field matrix
    # Singular values are stored in vector s
    # Left vectors are columns of U
    # Right vectors are columns of V
    
    s = np.load('singularValues.npy')
    U = np.load('leftVectors.npy')
    V = np.load('rightVectors.npy')
    
    #==============================================================================
    # Compute the Tikhonov-regularized solution to the near-field equation N * phi = TF.
    # 'TF' is a test function
    # 'alpha' is the optimal value of the regularization parameter (computed by
    # Morozov's discrepancy principle)
    # 'phi_alpha' is the regularized solution given 'alpha'
    
    datadir = np.load('datadir.npz')
    recordingTimes = np.load(str(datadir['recordingTimes']))
    receiverPoints = np.load(str(datadir['receivers']))
    
    # compute length of time step
    dt = (recordingTimes[-1] - recordingTimes[0]) / (len(recordingTimes) - 1)
    
    if Path('window.npz').exists():
        windowDict = np.load('window.npz')
        
        # Apply the time window
        tstart = windowDict['tstart']
        tstop = windowDict['tstop']
        tstep = windowDict['tstep']
        
        twStart = int(round(tstart / dt))
        twStop = int(round(tstop / dt))
        
        # Apply the receiver window
        rstart = windowDict['rstart']
        rstop = windowDict['rstop']
        rstep = windowDict['rstep']
        
    else:
        tstart = recordingTimes[0]
        twStart = 0
        tstop = recordingTimes[-1]
        twStop = len(recordingTimes)
        tstep = 1
        
        rstart = 0
        rstop = receiverPoints.shape[0]
        rstep = 1
        
    tinterval = np.arange(twStart, twStop, tstep)
    timeWindow = recordingTimes[tinterval]
        
    rinterval = np.arange(rstart, rstop, rstep)
    receiverPoints = receiverPoints[rinterval, :]    
    
    Nr = receiverPoints.shape[0]
    Nt = len(recordingTimes)
    Nw = len(timeWindow)
    
    velocity = pulseFun.velocity
    peakFreq = pulseFun.peakFreq
    peakTime = pulseFun.peakTime
    
    eps = np.finfo(float).eps     # about 2e-16 (so we never divide by zero)
    
    grid = np.load('grid.npz')
    if int(grid['ndim']) == 2:
        x = grid['x']
        y = grid['y']
        X, Y = np.meshgrid(x, y, indexing='ij')
        
        # Initialize the indicator function
        indicator = np.zeros(X.shape)
        
        if medium == 'constant':   
            pulse = lambda t : pulseFun.pulse(t)
            
            if Path('VZTestFuncsLSM.npz').exists():
                print('\nDetected that free-space test functions have already been simulated...')
                print('Checking consistency with current sampling grid and source-time function...')
                TFDict = np.load('VZTestFuncsLSM.npz')
                
                if np.array_equal(TFDict['x'], x) and np.array_equal(TFDict['y'], y) and np.array_equal(TFDict['time'], timeWindow):
                    print('Sampling grid is consistent...')
                    if TFDict['peakFreq'] == peakFreq and TFDict['peakTime'] == peakTime and TFDict['velocity'] == velocity:
                        print('Source-time function is consistent...')
                        if TFDict['tau'] == tau:
                            print('Source time %s is consistent...' %(tau))
                            print('Moving forward to imaging algorithm...')
                            TFshifted = TFDict['TFshifted']
                            TF = TFshifted[:, tinterval, :]
                            TFarray = TFDict['TFarray']
                            samplingPoints = TFDict['samplingPoints']
                            
                        else:
                            print('Current source time %s is inconsistent...' %(tau))
                            print('Shifting test functions to source time %s...' %(tau))
                            TFarray = TFDict['TFarray']
                            TFshifted = timeShift(TFarray, tau, dt)
                            TF = TFshifted[:, tinterval, :]
                            samplingPoints = TFDict['samplingPoints']
                    
                    elif TFDict['peakFreq'] != peakFreq or TFDict['peakTime'] != peakTime or TFDict['velocity'] != velocity:
                        print('Current source-time function is inconsistent...')
                        print('Resimulating test functions...')
                        samplingPoints = np.zeros((len(x) * len(y), 2))
                        TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                        
                        k = 0
                        for ix in trange(len(x)):
                            for iy in range(len(y)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy]])
                                TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                                       velocity, receiverPoints, samplingPoints[k, :])
                                k += 1
                                sleep(0.001)
                    
                        print('Shifting test functions to source time %s...' %(tau))
                        TFshifted = timeShift(TFarray, tau, dt)
                        TF = TFshifted[:, tinterval, :]
                
                else:
                    print('Current sampling grid is inconsistent...')
                    print('Resimulating test functions...')
                    # Initialize arrays for sampling points and test functions
                    samplingPoints = np.zeros((len(x) * len(y), 2))
                    TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                    
                    k = 0
                    for ix in trange(len(x)):
                        for iy in range(len(y)):
                            samplingPoints[k, :] = np.asarray([x[ix], y[iy]])
                            TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                                   velocity, receiverPoints, samplingPoints[k, :])
                            k += 1
                            sleep(0.001)
                    
                    print('Shifting test functions to source time %s...' %(tau))
                    TFshifted = timeShift(TFarray, tau, dt)
                    TF = TFshifted[:, tinterval, :]
                
            else:
                # Initialize arrays for sampling points and test functions
                samplingPoints = np.zeros((len(x) * len(y), 2))
                TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                
                print('\nSimulating free-space test functions for the current sampling grid...')
                k = 0
                for ix in trange(len(x)):
                    for iy in range(len(y)):
                        samplingPoints[k, :] = np.asarray([x[ix], y[iy]])
                        TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                               velocity, receiverPoints, samplingPoints[k, :])
                        k += 1
                        sleep(0.001)
                
                print('Shifting test functions to source time %s...' %(tau))
                TFshifted = timeShift(TFarray, tau, dt)
                TF = TFshifted[:, tinterval, :]
            
            np.savez('VZTestFuncsLSM.npz', TFarray=TFarray, TFshifted=TFshifted,
                     time=timeWindow, tinterval=tinterval, tau=tau,
                     peakFreq=peakFreq, peakTime=peakTime, velocity=velocity,
                     x=x, y=y, samplingPoints=samplingPoints)
            
            print('Reconstructing the scatterer...')
            k = 0
            for ix in trange(len(x), desc='Sampling'):
                for iy in range(len(y)):
                    tf = np.reshape(TF[:, :, k], (Nw * Nr, 1))
                    phi_alpha = Tikhonov(U, s, V, tf, alpha)
                    indicator[ix, iy] = 1.0 / (norm(phi_alpha) + eps)
                    k += 1
                    sleep(0.001)
            
        elif medium == 'variable':
            if 'testFuncs' in datadir:
                TFarray = np.load(str(datadir['testFuncs']))
                
            else:
                sys.exit('''
                         FileNotFoundError: Attempted to load file containing test
                         functions, but no such file exists. If a file exists containing
                         the test functions, run 'vzdata --path=<path/to/data/>' command 
                         and specify the file containing the test functions when prompted.
                         Otherwise, specify 'no' when asked if a file containing the test
                         functions exists.
                         ''')
            
            samplingPoints = np.zeros((len(x) * len(y), 2))
            userResponded = False
            print('''
                  In what order was the sampling grid spanned to compute the test functions?
                  Enter 'xy' if for each x, loop over y. (Default)
                  Enter 'yx' if for each y, loop over x.
                  Enter 'q/quit' to abort the calculation.
                  ''')
            while userResponded == False:
                order = input('Order: ')
                if order == '' or order == 'xy':
                    print('Proceeding with order \'xy\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for ix in trange(len(x), desc='Sampling'):
                        for iy in range(len(y)):
                            samplingPoints[k, :] = np.asarray([x[ix], y[iy]])
                            TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                            phi_alpha = Tikhonov(U, s, V, TF, alpha)
                            indicator[ix, iy] = 1.0 / (norm(phi_alpha) + eps)
                            k += 1
                            sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'yx':
                    print('Proceeding with order \'yx\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for iy in trange(len(y), desc='Sampling'):
                        for ix in range(len(x)):
                            samplingPoints[k, :] = np.asarray([x[ix], y[iy]])
                            TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                            phi_alpha = Tikhonov(U, s, V, TF, alpha)
                            indicator[ix, iy] = 1.0 / (norm(phi_alpha) + eps)
                            k += 1
                            sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'q' or order == 'quit':
                    sys.exit('Aborting calculation.')
                
                else:
                    print('''
                          Invalid response. Please enter one of the following:
                          Enter 'xy' if for each x, loop over y. (Default)
                          Enter 'yx' if for each y, loop over x.
                          Enter 'q/quit' to abort the calculation.
                          ''')
                        
            np.savez('VZTestFuncsLSM.npz', TFarray=TFarray, TFshifted=TFshifted,
                     time=timeWindow, tinterval=tinterval, tau=tau,
                     peakFreq=peakFreq, peakTime=peakTime, velocity=velocity,
                     x=x, y=y, samplingPoints=samplingPoints)
            
        np.savez('imageLSM.npz', ndim=2, indicator=indicator, alpha=alpha, tau=tau, X=X, Y=Y)
        
    elif int(grid['ndim']) == 3:
        x = grid['x']
        y = grid['y']
        z = grid['z']
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        
        # Initialize the indicator function
        indicator = np.zeros(X.shape)
        
        if medium == 'constant': 
            pulse = lambda t : pulseFun.pulse(t)
            
            if Path('VZTestFuncsLSM.npz').exists():
                print('\nDetected that free-space test functions have already been simulated...')
                print('Checking consistency with current sampling grid and source-time function...')
                TFDict = np.load('VZTestFuncsLSM.npz')
                
                if np.array_equal(TFDict['x'], x) and np.array_equal(TFDict['y'], y) and np.array_equal(TFDict['z'], z) and np.array_equal(TFDict['time'], timeWindow):
                    print('Sampling grid is consistent...')
                    if TFDict['peakFreq'] == peakFreq and TFDict['peakTime'] == peakTime and TFDict['velocity'] == velocity:
                        print('Source-time function is consistent...')
                        if TFDict['tau'] == tau:
                            print('Source time %s is consistent...' %(tau))
                            print('Moving forward to imaging algorithm...')
                            TFshifted = TFDict['TFshifted']
                            TF = TFshifted[:, tinterval, :]
                            TFarray = TFDict['TFarray']
                            samplingPoints = TFDict['samplingPoints']
                            
                        else:
                            print('Current source time %s is inconsistent...' %(tau))
                            print('Shifting test functions to source time %s...' %(tau))
                            TFarray = TFDict['TFarray']
                            TFshifted = timeShift(TFarray, tau, dt)
                            TF = TFshifted[:, tinterval, :]
                            samplingPoints = TFDict['samplingPoints']
                    
                    elif TFDict['peakFreq'] != peakFreq or TFDict['peakTime'] != peakTime or TFDict['velocity'] != velocity:
                        print('Current source-time function is inconsistent...')
                        print('Resimulating test functions...')
                        samplingPoints = np.zeros((len(x) * len(y) * len(z), 3))
                        TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                        
                        k = 0
                        for ix in trange(len(x)):
                            for iy in range(len(y)):
                                for iz in range(len(z)):
                                    samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                    TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                                           velocity, receiverPoints, samplingPoints[k, :])
                                    k += 1
                                    sleep(0.001)
                    
                        print('Shifting test functions to source time %s...' %(tau))
                        TFshifted = timeShift(TFarray, tau, dt)
                        TF = TFshifted[:, tinterval, :]
                
                else:
                    print('Current sampling grid is inconsistent...')
                    print('Resimulating test functions...')
                    # Initialize arrays for sampling points and test functions
                    samplingPoints = np.zeros((len(x) * len(y) * len(z), 3))
                    TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                    
                    k = 0
                    for ix in trange(len(x)):
                        for iy in range(len(y)):
                            for iz in range(len(z)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                                       velocity, receiverPoints, samplingPoints[k, :])
                                k += 1
                                sleep(0.001)
                    
                    print('Shifting test functions to source time %s...' %(tau))
                    TFshifted = timeShift(TFarray, tau, dt)
                    TF = TFshifted[:, tinterval, :]
                
            else:
                # Initialize arrays for sampling points and test functions
                samplingPoints = np.zeros((len(x) * len(y) * len(z), 3))
                TFarray = np.zeros((Nr, Nt, samplingPoints.shape[0]))
                
                print('\nSimulating free-space test functions for the current sampling grid...')
                k = 0
                for ix in trange(len(x)):
                    for iy in range(len(y)):
                        for iz in range(len(z)):
                            samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                            TFarray[:, :, k] = FundSol(pulse, recordingTimes,
                                   velocity, receiverPoints, samplingPoints[k, :])
                            k += 1
                            sleep(0.001)
                
                print('Shifting test functions to source time %s...' %(tau))
                TFshifted = timeShift(TFarray, tau, dt)
                TF = TFshifted[:, tinterval, :]
            
            np.savez('VZTestFuncsLSM.npz', TFarray=TFarray, TFshifted=TFshifted,
                     time=timeWindow, tinterval=tinterval, tau=tau,
                     peakFreq=peakFreq, peakTime=peakTime, velocity=velocity,
                     x=x, y=y, z=z, samplingPoints=samplingPoints)
            
            print('Reconstructing the scatterer...')
            k = 0
            for ix in trange(len(x), desc='Sampling'):
                for iy in range(len(y)):
                    for iz in range(len(z)):
                        samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                        TFarray[:, :, k] = FundSol(pulse, recordingTimes - tau,
                              velocity, receiverPoints, samplingPoints[k, :])
                        TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                        phi_alpha = Tikhonov(U, s, V, TF, alpha)
                        indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                        k += 1
                        sleep(0.001)
                        
        elif medium == 'variable':
            if 'testFuncs' in datadir:
                TFarray = np.load(str(datadir['testFuncs']))
            else:
                sys.exit('''
                         FileNotFoundError: Attempted to load the file containing the test
                         functions, but no such file exists. If a file exists containing
                         the test functions, run 'vzdata --path=<path/to/data/>' command 
                         and specify the file containing the test functions when prompted.
                         Otherwise, specify 'no' when asked if a file containing the test
                         functions exists.
                         ''')
            TFarray = TFarray[rinterval, :, :]
            if tau != 0:
                TFarray = timeShift(TFarray, tau, dt)
            TFarray = TFarray[:, tinterval, :]
            
            samplingPoints = np.zeros((len(x) * len(y) * len(z), 3))
            userResponded = False
            print('''
                  In what order was the sampling grid spanned to compute the test functions?
                  Enter 'xyz' if for each x, for each y, loop over z. (Default)
                  Enter 'xzy' if for each x, for each z, loop over y.
                  Enter 'yxz' if for each y, for each x, loop over z.
                  Enter 'yzx' if for each y, for each z, loop over x.
                  Enter 'zxy' if for each z, for each x, loop over y.
                  Enter 'zyx' if for each z, for each y, loop over x
                  Enter 'q/quit' to abort the calculation.
                  ''')
            while userResponded == False:
                order = input('Order: ')
                if order == '' or order == 'xyz':
                    print('Proceeding with order \'xyz\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for ix in trange(len(x), desc='Sampling'):
                        for iy in range(len(y)):
                            for iz in range(len(z)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'xzy':
                    print('Proceeding with order \'xzy\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for ix in trange(len(x), desc='Sampling'):
                        for iz in range(len(z)):
                            for iy in range(len(y)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'yxz':
                    print('Proceeding with order \'yxz\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for iy in trange(len(y), desc='Sampling'):
                        for ix in range(len(x)):
                            for iz in range(len(z)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'yzx':
                    print('Proceeding with order \'yzx\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for iy in trange(len(y), desc='Sampling'):
                        for iz in range(len(z)):
                            for ix in range(len(x)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'zxy':
                    print('Proceeding with order \'zxy\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for iz in trange(len(z), desc='Sampling'):
                        for ix in range(len(x)):
                            for iy in range(len(y)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'zyx':
                    print('Proceeding with order \'zyx\'...')
                    print('Reconstructing the scatterer...')
                    k = 0
                    for iz in trange(len(z), desc='Sampling'):
                        for iy in range(len(y)):
                            for ix in range(len(x)):
                                samplingPoints[k, :] = np.asarray([x[ix], y[iy], z[iz]])
                                TF = np.reshape(TFarray[:, :, k], (Nt * Nr, 1))
                                phi_alpha = Tikhonov(U, s, V, TF, alpha)
                                indicator[ix, iy, iz] = 1.0 / (norm(phi_alpha) + eps)
                                k += 1
                                sleep(0.001)
                    
                    userResponded = True
                    break
                
                elif order == 'q' or order == 'quit':
                    sys.exit('Aborting calculation.')
                else:
                    print('''
                          Invalid response. Please enter one of the following:
                          Enter 'xyz' if for each x, for each y, loop over z. (Default)
                          Enter 'xzy' if for each x, for each z, loop over y.
                          Enter 'yxz' if for each y, for each x, loop over z.
                          Enter 'yzx' if for each y, for each z, loop over x.
                          Enter 'zxy' if for each z, for each x, loop over y.
                          Enter 'zyx' if for each z, for each y, loop over x
                          Enter 'q/quit' to abort the calculation.
                          ''')
                        
            np.savez('VZTestFuncsLSM.npz', TFarray=TFarray, TFshifted=TFshifted,
                     time=timeWindow, tinterval=tinterval, tau=tau,
                     peakFreq=peakFreq, peakTime=peakTime, velocity=velocity,
                     x=x, y=y, z=z, samplingPoints=samplingPoints)
        
        np.savez('imageLSM.npz', ndim=3, indicator=indicator, alpha=alpha, tau=tau, X=X, Y=Y, Z=Z)