# Copyright 2017-2018 Aaron C. Prunty
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#        
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#==============================================================================

import sys
import argparse
import pickle
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.ticker import FormatStrFormatter
from mpl_toolkits.axes_grid1 import make_axes_locatable
from skimage import measure
from pathlib import Path

def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument('--method', type=str, choices=['lsm', 'telsm'],
                        help='''Specify whether to plot the test image from the classical
                        linear sampling method 'lsm' or total-energy linear sampling method 'telsm'.''')
    parser.add_argument('--format', '-f', type=str, default=None, choices=['png', 'pdf', 'ps', 'eps', 'svg'],
                        help='''specify the image format of the saved file. Accepted formats are png, pdf,
                        ps, eps, and svg. Default format is set to pdf.''')
    parser.add_argument('--xlabel', type=str, default=None,
                        help='''specify a label for the x-axis.''')
    parser.add_argument('--ylabel', type=str, default=None,
                        help='''specify a label for the y-axis.''')
    parser.add_argument('--zlabel', type=str, default=None,
                        help='''specify a label for the z-axis.''')
    parser.add_argument('--xu', type=str, default=None,
                        help='Specify the units for the x-axis (e.g., \'m\' or \'km\').')
    parser.add_argument('--yu', type=str, default=None,
                        help='Specify the units for the y-axis (e.g., \'m\' or \'km\').')
    parser.add_argument('--zu', type=str, default=None,
                        help='Specify the units for the z-axis (e.g., \'m\' or \'km\').')
    parser.add_argument('--colorbar', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'y/yes/true\' to plot colorbar. Default is \'n/no/false\'')
    parser.add_argument('--invert_xaxis', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'y/yes/true\' to invert x-axis. Default is \'n/no/false\'')
    parser.add_argument('--invert_yaxis', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'y/yes/true\' to invert y-axis. Default is \'n/no/false\'')
    parser.add_argument('--invert_zaxis', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'y/yes/true\' to invert z-axis. Default is \'n/no/false\'')
    parser.add_argument('--show_scatterer', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'y/yes/true\' to show scatterer boundary. Default is \'n/no/false\'')
    parser.add_argument('--show_sources', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'n/no/false\' to hide sources. Default is \'y/yes/true\'')
    parser.add_argument('--show_receivers', type=str, default=None, choices=['n', 'no', 'false', 'y', 'yes', 'true'],
                        help='specify \'n/no/false\' to hide receivers. Default is \'y/yes/true\'')
    args = parser.parse_args()
    
    # if a plotParams.pkl file already exists, load relevant parameters
    if Path('plotParams.pkl').exists():
        plotParams = pickle.load(open('plotParams.pkl', 'rb'))
        
        # for both wiggle plots and image plots
        if args.format is not None:
            pltformat = args.format
            plotParams['pltformat'] = pltformat
        else:
            pltformat = plotParams['pltformat']
        
        # for image/map plots
        if args.xlabel is not None:
            xlabel = args.xlabel
            plotParams['xlabel'] = xlabel
        else:
            xlabel = plotParams['xlabel']
        
        if args.ylabel is not None:
            ylabel = args.ylabel
            plotParams['ylabel'] = ylabel
        else:
            ylabel = plotParams['ylabel']
            
        if args.zlabel is not None:
            zlabel = args.zlabel
            plotParams['zlabel'] = zlabel
        else:
            xlabel = plotParams['xlabel']            
        #==============================================================================
        # handle units here
        if all(v is not None for v in [args.xu, args.yu, args.zu]):
            xu = args.xu
            yu = args.yu
            zu = args.zu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.xu is not None and all(v is None for v in [args.yu, args.zu]):
            xu = args.xu
            yu = xu
            zu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.yu is not None and all(v is None for v in [args.xu, args.zu]):
            yu = args.yu
            xu = yu
            zu = yu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.zu is not None and all(v is None for v in [args.xu, args.yu]):
            zu = args.zu
            xu = zu
            yu = zu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.zu is None and all(v is not None for v in [args.xu, args.yu]):
            xu = args.xu
            yu = args.yu
            zu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.yu is None and all(v is not None for v in [args.xu, args.zu]):
            xu = args.xu
            zu = args.zu
            yu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.xu is None and all(v is not None for v in [args.yu, args.zu]):
            yu = args.yu
            zu = args.zu
            xu = yu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif all(v is None for v in [args.xu, args.yu, args.zu]):
            xu = plotParams['xu']
            yu = plotParams['yu']
            zu = plotParams['zu']
        #==============================================================================    
        if args.colorbar is not None:
            if args.colorbar == 'n' or args.colorbar == 'no' or args.colorbar == 'false':
                wantColorbar = False
                if plotParams['colorbar'] is True:
                    plotParams['colorbar'] = False
            elif args.colorbar == 'y' or args.colorbar == 'yes' or args.colorbar == 'true':
                wantColorbar = True
                if plotParams['colorbar'] is False:
                    plotParams['colorbar'] = True
        else:
            wantColorbar = plotParams['colorbar']
        #==============================================================================    
        if args.invert_xaxis is not None:
            if args.invert_xaxis == 'n' or args.invert_xaxis == 'no' or args.invert_xaxis == 'false':
                invertX = False
                if plotParams['invert_xaxis'] is True:
                    plotParams['invert_xaxis'] = False
            elif args.invert_xaxis == 'y' or args.invert_xaxis == 'yes' or args.invert_xaxis == 'true':
                invertX = True
                if plotParams['invert_xaxis'] is False:
                    plotParams['invert_xaxis'] = True
        else:
            invertX = plotParams['invert_xaxis']
        #==============================================================================
        if args.invert_yaxis is not None:
            if args.invert_yaxis == 'n' or args.invert_yaxis == 'no' or args.invert_yaxis == 'false':
                invertY = False
                if plotParams['invert_yaxis'] is True:
                    plotParams['invert_yaxis'] = False
            elif args.invert_yaxis == 'y' or args.invert_yaxis == 'yes' or args.invert_yaxis == 'true':
                invertY = True
                if plotParams['invert_yaxis'] is False:
                    plotParams['invert_yaxis'] = True
        else:
            invertY = plotParams['invert_yaxis']
        #==============================================================================
        if args.invert_zaxis is not None:
            if args.invert_zaxis == 'n' or args.invert_zaxis == 'no' or args.invert_zaxis == 'false':
                invertZ = False
                if plotParams['invert_zaxis'] is True:
                    plotParams['invert_zaxis'] = False
            elif args.invert_zaxis == 'y' or args.invert_zaxis == 'yes' or args.invert_zaxis == 'true':
                invertZ = True
                if plotParams['invert_zaxis'] is False:
                    plotParams['invert_zaxis'] = True
        else:
            invertZ = plotParams['invert_zaxis']
        #==============================================================================
        if args.show_scatterer is not None:
            if args.show_scatterer == 'n' or args.show_scatterer == 'no' or args.show_scatterer == 'false':
                showScatterer = False
                if plotParams['show_scatterer'] is True:
                    plotParams['show_scatterer'] = False
            elif args.show_scatterer == 'y' or args.show_scatterer == 'yes' or args.show_scatterer == 'true':
                showScatterer = True
                if plotParams['show_scatterer'] is False:
                    plotParams['show_scatterer'] = True
        else:
            showScatterer = plotParams['show_scatterer']
        #==============================================================================
        if args.show_sources is not None:
            if args.show_sources == 'n' or args.show_sources == 'no' or args.show_sources == 'false':
                showSources = False
                if plotParams['show_sources'] is True:
                    plotParams['show_sources'] = False
            elif args.show_sources == 'y' or args.show_sources == 'yes' or args.show_sources == 'true':
                showSources = True
                if plotParams['show_sources'] is False:
                    plotParams['show_sources'] = True
        else:
            showSources = plotParams['show_sources']
        #==============================================================================
        if args.show_receivers is not None:
            if args.show_receivers == 'n' or args.show_receivers == 'no' or args.show_receivers == 'false':
                showReceivers = False
                if plotParams['show_receivers'] is True:
                    plotParams['show_receivers'] = False
            elif args.show_receivers == 'y' or args.show_receivers == 'yes' or args.show_receivers == 'true':
                showReceivers = True
                if plotParams['show_receivers'] is False:
                    plotParams['show_receivers'] = True
        else:
            showReceivers = plotParams['show_receivers']
        #==============================================================================
        tu = plotParams['tu']
        
        pickle.dump(plotParams, open('plotParams.pkl', 'wb'), pickle.HIGHEST_PROTOCOL)
    #==============================================================================
    else: # create a plotParams dictionary file with default values
        plotParams = {}
        #for both image and wiggle plots
        if args.format is None:
            pltformat = 'pdf'
        else:
            pltformat = args.format
        plotParams['pltformat'] = pltformat
        
        # for image/map plots
        if args.colorbar is not None:
            if args.colorbar == 'n' or args.colorbar == 'no' or args.colorbar == 'false':
                wantColorbar = False
                plotParams['colorbar'] = wantColorbar
            elif args.colorbar == 'y' or args.colorbar == 'yes' or args.colorbar == 'true':
                wantColorbar = True
                plotParams['colorbar'] = wantColorbar
        else:
            wantColorbar = False
            plotParams['colorbar'] = wantColorbar
        
        if args.xlabel is None:
            xlabel = ''
        else:
            xlabel = args.xlabel
        plotParams['xlabel'] = xlabel
        
        if args.ylabel is None:
            ylabel = ''
        else:
            ylabel = args.ylabel
        plotParams['ylabel'] = ylabel
        
        if args.zlabel is None:
            zlabel = ''
        else:
            zlabel = args.zlabel
        plotParams['zlabel'] = zlabel
        #==============================================================================
        if all(v is not None for v in [args.xu, args.yu, args.zu]):
            xu = args.xu
            yu = args.yu
            zu = args.zu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.xu is not None and all(v is None for v in [args.yu, args.zu]):
            xu = args.xu
            yu = xu
            zu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.yu is not None and all(v is None for v in [args.xu, args.zu]):
            yu = args.yu
            xu = yu
            zu = yu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.zu is not None and all(v is None for v in [args.xu, args.yu]):
            zu = args.zu
            xu = zu
            yu = zu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.zu is None and all(v is not None for v in [args.xu, args.yu]):
            xu = args.xu
            yu = args.yu
            zu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.yu is None and all(v is not None for v in [args.xu, args.zu]):
            xu = args.xu
            zu = args.zu
            yu = xu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif args.xu is None and all(v is not None for v in [args.yu, args.zu]):
            yu = args.yu
            zu = args.zu
            xu = yu
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        elif all(v is None for v in [args.xu, args.yu, args.zu]):
            xu = ''
            yu = ''
            zu = ''
            plotParams['xu'] = xu
            plotParams['yu'] = yu
            plotParams['zu'] = zu
        #==============================================================================
        if args.invert_xaxis is not None:
            if args.invert_xaxis == 'n' or args.invert_xaxis == 'no' or args.invert_xaxis == 'false':
                invertX = False
                plotParams['invert_xaxis'] = invertX
            elif args.invert_xaxis == 'y' or args.invert_xaxis == 'yes' or args.invert_xaxis == 'true':
                invertX = True
                plotParams['invert_xaxis'] = invertX
        else:
            invertX = False
            plotParams['invert_xaxis'] = invertX
        #==============================================================================    
        if args.invert_yaxis is not None:
            if args.invert_yaxis == 'n' or args.invert_yaxis == 'no' or args.invert_yaxis == 'false':
                invertY = False
                plotParams['invert_yaxis'] = invertY
            elif args.invert_yaxis == 'y' or args.invert_yaxis == 'yes' or args.invert_yaxis == 'true':
                invertY = True
                plotParams['invert_yaxis'] = invertY
        else:
            invertY = False
            plotParams['invert_yaxis'] = invertY
        #==============================================================================    
        if args.invert_zaxis is not None:
            if args.invert_zaxis == 'n' or args.invert_zaxis == 'no' or args.invert_zaxis == 'false':
                invertZ = False
                plotParams['invert_zaxis'] = invertZ
            elif args.invert_zaxis == 'y' or args.invert_zaxis == 'yes' or args.invert_zaxis == 'true':
                invertZ = True
                plotParams['invert_zaxis'] = invertZ
        else:
            invertZ = False
            plotParams['invert_zaxis'] = invertZ
        #==============================================================================
        if args.show_scatterer is not None:
            if args.show_scatterer == 'n' or args.show_scatterer == 'no' or args.show_scatterer == 'false':
                showScatterer = False
                plotParams['show_scatterer'] = showScatterer
            elif args.show_scatterer == 'y' or args.show_scatterer == 'yes' or args.show_scatterer == 'true':
                showScatterer = True
                plotParams['show_scatterer'] = showScatterer
        else:
            showScatterer = False
            plotParams['show_scatterer'] = showScatterer
        #==============================================================================
        if args.show_sources is not None:
            if args.show_sources == 'n' or args.show_sources == 'no' or args.show_sources == 'false':
                showSources = False
                plotParams['show_sources'] = showSources
            elif args.show_sources == 'y' or args.show_sources == 'yes' or args.show_sources == 'true':
                showSources = True
                plotParams['show_sources'] = showSources
        else:
            showSources = True
            plotParams['show_sources'] = showSources
        #==============================================================================
        if args.show_receivers is not None:
            if args.show_receivers == 'n' or args.show_receivers == 'no' or args.show_receivers == 'false':
                showReceivers = False
                plotParams['show_receivers'] = showReceivers
            elif args.show_receivers == 'y' or args.show_receivers == 'yes' or args.show_receivers == 'true':
                showReceivers = True
                plotParams['show_receivers'] = showReceivers
        else:
            showReceivers = True
            plotParams['show_receivers'] = showReceivers
        #==============================================================================
        tu = ''
        plotParams['tu'] = tu
        au = ''
        plotParams['au'] = au
        
        pickle.dump(plotParams, open('plotParams.pkl', 'wb'), pickle.HIGHEST_PROTOCOL)
    
    #==============================================================================
    # load the shape of the scatterer and source/receiver locations
    datadir = np.load('datadir.npz')
    receiverPoints = np.load(str(datadir['receivers']))
    sourcePoints = np.load(str(datadir['sources']))
    if 'scatterer' in datadir:
        scatterer = np.load(str(datadir['scatterer']))
    else:
        scatterer = None
        
    if scatterer is None and showScatterer == True:
        sys.exit('''
              Attempted to load the file containing the scatterer coordinates,
              but no such file exists. If a file exists containing the scatterer
              points, run 'vzdata --path=<path/to/data/>' command and specify
              the file containing the scatterer points when prompted. Otherwise,
              specify 'no' when asked if a file containing the scatterer points
              exists.
              ''')
    
    if Path('window.npz').exists():
        windowDict = np.load('window.npz')
        
        # Apply the receiver window
        rstart = windowDict['rstart']
        rstop = windowDict['rstop']
        rstep = windowDict['rstep']
        
        rinterval = np.arange(rstart, rstop, rstep)
        receiverPoints = receiverPoints[rinterval, :]
        
        # Apply the source window
        sstart = windowDict['sstart']
        sstop = windowDict['sstop']
        sstep = windowDict['sstep']
        
        sinterval = np.arange(sstart, sstop, sstep)
        sourcePoints = sourcePoints[sinterval, :]
    #==============================================================================
    if Path('imageLSM.npz').exists() and not Path('imageTELSM.npz').exists():
        if args.method is None or args.method == 'lsm':
            LSMDict = np.load('imageLSM.npz')
            indicator = LSMDict['indicator']
            alpha = LSMDict['alpha']
            tau = LSMDict['tau']
            
            if int(LSMDict['ndim']) == 2:
                X = LSMDict['X']
                Y = LSMDict['Y']
                
                fig = plt.figure()
                ax = fig.add_subplot(111)
                im = ax.contourf(X, Y, indicator, 100, cmap=plt.cm.magma)
                    
            elif int(LSMDict['ndim']) == 3:
                X = LSMDict['X']
                Y = LSMDict['Y']
                #Z = LSMDict['Z']
                
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                verts, faces, normals, values = measure.marching_cubes_lewiner(indicator)
                ax.plot_trisurf(verts[:, 0], verts[:,1], faces, verts[:, 2], color='silver')
        
            if alpha != 0 and tu != '':
                ax.set_title(r'$\alpha = %0.1e, \tau = %s$ %s' %(alpha, tau, tu))
            elif alpha != 0 and tu == '':
                ax.set_title(r'$\alpha = %0.1e, \tau = %s$' %(alpha, tau))
            elif alpha == 0 and tu != '':
                ax.set_title(r'$\alpha = %s, \tau = %s$ %s' %(alpha, tau, tu))
            elif alpha == 0 and tu == '':
                ax.set_title(r'$\alpha = %s, \tau = %s$' %(alpha, tau))
            
            if wantColorbar:
                divider = make_axes_locatable(ax)
                cax = divider.append_axes('right', size='5%', pad=0.05)
                cbar = plt.colorbar(im, cax=cax)
                cbar.ax.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
                cbar.formatter.set_scientific(True)
                cbar.formatter.set_powerlimits((0, 0))
                cbar.set_label(r'$\frac{1}{\vert\vert\varphi\vert\vert}$', labelpad=24, rotation=0, fontsize=18)
                
    elif not Path('imageLSM.npz').exists() and Path('imageTELSM.npz').exists():
        if args.method is None or args.method == 'telsm':
            TELSMDict = np.load('imageTELSM.npz')
            Image = TELSMDict['Image']
            alpha = TELSMDict['alpha']
            
            if int(TELSMDict['ndim']) == 2:
                X = TELSMDict['X']
                Y = TELSMDict['Y']
                
                fig = plt.figure()
                ax = fig.add_subplot(111)
                im = ax.contourf(X, Y, Image, 100, cmap=plt.cm.magma)
                    
            elif int(TELSMDict['ndim']) == 3:
                X = TELSMDict['X']
                Y = TELSMDict['Y']
                #Z = TELSMDict['Z']
                
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                verts, faces, normals, values = measure.marching_cubes_lewiner(Image)
                ax.plot_trisurf(verts[:, 0], verts[:,1], faces, verts[:, 2], color='silver')
        
            if alpha != 0:
                ax.set_title(r'$\alpha = %0.1e$' %(alpha))
            else:
                ax.set_title(r'$\alpha = %s$' %(alpha))
            
            if wantColorbar:
                divider = make_axes_locatable(ax)
                cax = divider.append_axes('right', size='5%', pad=0.05)
                cbar = plt.colorbar(im, cax=cax)
                
    elif Path('imageLSM.npz').exists() and Path('imageTELSM.npz').exists():
        if args.method == 'lsm':
            LSMDict = np.load('imageLSM.npz')
            indicator = LSMDict['indicator']
            alpha = LSMDict['alpha']
            tau = LSMDict['tau']
            
            if int(LSMDict['ndim']) == 2:
                X = LSMDict['X']
                Y = LSMDict['Y']
                
                fig = plt.figure()
                ax = fig.add_subplot(111)
                im = ax.contourf(X, Y, indicator, 100, cmap=plt.cm.magma)
                    
            elif int(LSMDict['ndim']) == 3:
                X = LSMDict['X']
                Y = LSMDict['Y']
                #Z = LSMDict['Z']
                
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                verts, faces, normals, values = measure.marching_cubes_lewiner(indicator)
                ax.plot_trisurf(verts[:, 0], verts[:,1], faces, verts[:, 2], color='silver')
        
            if alpha != 0 and tu != '':
                ax.set_title(r'$\alpha = %0.1e, \tau = %s$ %s' %(alpha, tau, tu))
            elif alpha != 0 and tu == '':
                ax.set_title(r'$\alpha = %0.1e, \tau = %s$' %(alpha, tau))
            elif alpha == 0 and tu != '':
                ax.set_title(r'$\alpha = %s, \tau = %s$ %s' %(alpha, tau, tu))
            elif alpha == 0 and tu == '':
                ax.set_title(r'$\alpha = %s, \tau = %s$' %(alpha, tau))
            
            if wantColorbar:
                divider = make_axes_locatable(ax)
                cax = divider.append_axes('right', size='5%', pad=0.05)
                cbar = plt.colorbar(im, cax=cax)
                cbar.ax.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
                cbar.formatter.set_scientific(True)
                cbar.formatter.set_powerlimits((0, 0))
                cbar.set_label(r'$\frac{1}{\vert\vert\varphi\vert\vert}$', labelpad=24, rotation=0, fontsize=18)
        
        elif args.method == 'telsm':
            TELSMDict = np.load('imageTELSM.npz')
            Image = TELSMDict['Image']
            alpha = TELSMDict['alpha']
            
            if int(TELSMDict['ndim']) == 2:
                X = TELSMDict['X']
                Y = TELSMDict['Y']
                
                fig = plt.figure()
                ax = fig.add_subplot(111)
                im = ax.contourf(X, Y, Image, 100, cmap=plt.cm.magma)
                    
            elif int(TELSMDict['ndim']) == 3:
                X = TELSMDict['X']
                Y = TELSMDict['Y']
                #Z = TELSMDict['Z']
                
                fig = plt.figure()
                ax = fig.add_subplot(111, projection='3d')
                verts, faces, normals, values = measure.marching_cubes_lewiner(Image)
                ax.plot_trisurf(verts[:, 0], verts[:,1], faces, verts[:, 2], color='silver')
        
            if alpha != 0 :
                ax.set_title(r'$\alpha = %0.1e$' %(alpha))
            else:
                ax.set_title(r'$\alpha = %s$' %(alpha))
            
            if wantColorbar:
                divider = make_axes_locatable(ax)
                cax = divider.append_axes('right', size='5%', pad=0.05)
                cbar = plt.colorbar(im, cax=cax)
    
    try:
        ax
    except NameError:
        ax = None
    
    if ax is None:
        if receiverPoints.shape[1] == 2:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            
            if showSources:
                ax.plot(sourcePoints[:, 0], sourcePoints[:, 1], 'k*')
            if showReceivers:
                ax.plot(receiverPoints[:, 0], receiverPoints[:, 1], 'kv')
            if scatterer is not None and showScatterer:
                ax.plot(scatterer[:, 0], scatterer[:, 1], '--', color='darkgray')
                
            if xu != '':
                ax.set_xlabel(xlabel + ' (%s)' %(xu))
            else:
                ax.set_xlabel(xlabel)

            if yu != '':
                ax.set_ylabel(ylabel + ' (%s)' %(yu))
            else:
                ax.set_ylabel(ylabel)
                
        elif receiverPoints.shape[1] == 3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            
            if showSources:
                ax.plot(sourcePoints[:, 0], sourcePoints[:, 1], sourcePoints[:, 2], 'k*')
            if showReceivers:
                ax.plot(receiverPoints[:, 0], receiverPoints[:, 1], receiverPoints[:, 2], 'kv')
            if scatterer is not None and showScatterer:
                ax.plot(scatterer[:, 0], scatterer[:, 1], scatterer[:, 2], '--', color='darkgray')
                
            if xu != '':
                ax.set_xlabel(xlabel + ' (%s)' %(xu))
            else:
                ax.set_xlabel(xlabel)

            if yu != '':
                ax.set_ylabel(ylabel + ' (%s)' %(yu))
            else:
                ax.set_ylabel(ylabel)
                
            if zu != '':
                ax.set_zlabel(zlabel + ' (%s)' %(zu))
            else:
                ax.set_zlabel(zlabel)
                
    else:
        if receiverPoints.shape[1] == 2:
            if showSources:
                ax.plot(sourcePoints[:, 0], sourcePoints[:, 1], 'k*')
            if showReceivers:
                ax.plot(receiverPoints[:, 0], receiverPoints[:, 1], 'kv')
            if scatterer is not None and showScatterer:
                ax.plot(scatterer[:, 0], scatterer[:, 1], '--', color='darkgray')
                
            if xu != '':
                ax.set_xlabel(xlabel + ' (%s)' %(xu))
            else:
                ax.set_xlabel(xlabel)

            if yu != '':
                ax.set_ylabel(ylabel + ' (%s)' %(yu))
            else:
                ax.set_ylabel(ylabel)
        
        elif receiverPoints.shape[1] == 3:
            if showSources:
                ax.plot(sourcePoints[:, 0], sourcePoints[:, 1], sourcePoints[:, 2], 'k*')
            if showReceivers:
                ax.plot(receiverPoints[:, 0], receiverPoints[:, 1], receiverPoints[:, 2], 'kv')
            if scatterer is not None and showScatterer:
                ax.plot(scatterer[:, 0], scatterer[:, 1], scatterer[:, 2], '--', color='darkgray')
            
            if xu != '':
                ax.set_xlabel(xlabel + ' (%s)' %(xu))
            else:
                ax.set_xlabel(xlabel)

            if yu != '':
                ax.set_ylabel(ylabel + ' (%s)' %(yu))
            else:
                ax.set_ylabel(ylabel)
                
            if zu != '':
                ax.set_zlabel(zlabel + ' (%s)' %(zu))
            else:
                ax.set_zlabel(zlabel)
    #==============================================================================
    ax.set_aspect('equal')
    if invertX:
        ax.invert_xaxis()
    if invertY:
        ax.invert_yaxis()
    if invertZ:
        ax.invert_zaxis()
        
    plt.tight_layout()
    plt.savefig('image.' + pltformat, format=pltformat, bbox_inches='tight', transparent=True)
    plt.show()
